import sys
import logging
import asyncio
import itertools

from starlette.applications import Starlette
from starlette.endpoints import HTTPEndpoint
from starlette.responses import Response, PlainTextResponse, StreamingResponse

import uvicorn
import aiohttp

from package.shared import models
from package.shared.logger import setup_logger
from package.shared.config import Config
from package.shared.util import board_cfg_file


app = Starlette(debug=False)
app.running = True


@app.on_event('startup')
async def startup_handler():
    pass


@app.on_event('shutdown')
async def close_app():
    global app
    logging.info('Shuting down application')
    app.running = False


@app.route('/')
class Homepage(HTTPEndpoint):
    async def get(self, request):
        return PlainTextResponse('Welcome')


@app.route('/mjpeg')
class Mjpeg(HTTPEndpoint):
    async def mjpeg_generator(self, stream):
        global app

        session = models.SessionSingleton.instance().session
        while app.running:
            buf = None

            try:
                async with aiohttp.ClientSession() as session:
                    async with session.get(stream.input_url) as response:
                        buf = await response.read()
            except Exception:
                yield ''
                return

            if buf is None:
                logging.warning('Failed to read mjpeg input')
                await asyncio.sleep(3)
                continue

            yield '--endofsection\r\n'
            yield 'Content-type: image/jpeg\r\n' + \
                'Content-Length: {}\r\n'.format(len(buf)) + \
                '\r\n'
            
            yield buf
            yield '\r\n'

            await asyncio.sleep(0.3)


    async def get(self, request):
        try:
            serial = request.query_params['stream']
        except Exception as e:
            return Response(str(e), status_code=404, headers=None, media_type=None)

        session = models.SessionSingleton.instance().session

        try:
            stream = session.query(models.Stream) \
                .filter(models.Stream.serial==serial) \
                .first()
        except Exception as e:
            logging.exception("Exception")
            session.rollback()
            return Response(str(e), status_code=500, headers=None, media_type=None)

        try:
            if not stream.klass.startswith('mjpeg'):
                raise Exception('Not a MJPEG stream')

            if not stream.is_forward:
                raise Exception('Not a proxy stream')

        except Exception as e:
            return Response('Stream not found', status_code=404, headers=None, media_type=None)

        return StreamingResponse(self.mjpeg_generator(stream), media_type='multipart/x-mixed-replace; boundary=endofsection')


def run_mjpeg(**kwargs):
    Config.init(board_cfg_file())
    setup_logger('mjpeg.txt')

    config = Config.get()
    models.SessionSingleton.recreate(config.app.database)

    uvicorn.run(
        app,
        host=config.mjpeg.host,
        port=int(config.mjpeg.port))


if __name__ == '__main__':
    run_mjpeg()
