from typing import Dict, List, Any

import sys
import asyncio
import logging
import datetime
import json
from dataclasses import dataclass

import zmq
from zmq.asyncio import Context as ZmqContext
from zmq.asyncio import Socket as ZmqSocket

from package.shared.constants import Constants
from package.shared import models
from package.shared.config import Config
from package.shared.logger import setup_logger
from package.shared.ipc import IPCWorker
from package.shared.flow_control import FlowControl
from package.shared.util import board_cfg_file


@dataclass
class DeviceInfo():
    task: asyncio.Task
    last_updated: datetime.datetime


@dataclass
class PosClient(IPCWorker):
    flow: FlowControl = None

    def __post_init__(self):
        super().__post_init__()


    def setup(self):
        config = Config.get()
        self.flow = FlowControl()
        self.m_task_by_name: Dict[DeviceInfo] = dict()

        # Zmq
        self.m_context = None
        self.m_pub = None

        self.setup_sockets()
        self.setup_ipc(url=config.app.ipc, ctx=self.m_context, type=zmq.DEALER, name='pos_client')


    def cleanup(self):
        self.close_sockets()


    def setup_sockets(self):
        config = Config.get()

        # Zmq
        self.m_context = ZmqContext()
        self.m_pub = self.m_context.socket(zmq.PUB)
        self.m_pub.setsockopt(zmq.LINGER, 500)
        self.m_pub.connect(config.bridge.internal)


    def close_sockets(self):
        self.close_ipc()
        self.m_pub.close()
        self.m_context.destroy()


    async def run_device(self, device: models.Device):
        config = Config.get()

        try:
            session = models.SessionSingleton.instance().session

            while True:
                reader = None
                writer = None

                ping_time = int(config.pos.ping)
                host = device.address
                port = device.port

                try:
                    reader, writer = await asyncio.open_connection(host, port)
                except:
                    reconnect_time = int(config.pos.reconnect)
                    logging.warning(f"Can't connect to pos device {host}:{port}, retry in {reconnect_time} secs")
                    await asyncio.sleep(reconnect_time)
                    continue

                logging.info(f"Connected to: {host}:{port}")

                async def read():
                    nonlocal reader
                    return await reader.readline()

                async def ping():
                    nonlocal writer
                    writer.write(b'\0')
                    await writer.drain()

                # Whole pos message
                pos_message = ''

                while True:
                    try:
                        data = await asyncio.wait_for(read(), timeout=ping_time)

                        if not data:
                            logging.warning(f"POS device {device.name} down")
                            break

                        line = data.decode()
                        line_strip = line.strip()
                        if len(line_strip) == 0:
                            pos_message = pos_message.strip()
                            if len(pos_message) > 0:
                                # Store
                                try:
                                    now = datetime.datetime.now()
                                    now_min = models.get_timestamp_min(now.timestamp())

                                    wholepos = models.WholePos(
                                        created_at=now,
                                        created_min=now_min,
                                        text=pos_message,
                                        device_id=device.id)

                                    session.add(wholepos)
                                    session.commit()
                                except:
                                    logging.exception(f"Save wholepos failed: {device.name}")
                                    session.rollback()

                                pos_message = ''
                        else:
                            pos_message = pos_message + line

                        await self.on_pos_received(line_strip, device)
                    except asyncio.TimeoutError:
                        await ping()
                    except asyncio.CancelledError:
                        break
                    except:
                        logging.exception(f"Exception: {device.name}")
                        break

                writer.close()
        except:
            logging.exception(f'run_device {device.name} exception')


    async def on_pos_received(self, line, device):
        logging.info(f"Received from {device.name}: {line}")

        created_at = datetime.datetime.now()
        created_min = models.get_timestamp_min(created_at.timestamp())

        session = models.SessionSingleton.instance().session
        try:
            pos = models.Pos(
                created_at=created_at,
                created_min=created_min,
                text=line,
                device=device)

            session.add(pos)
            session.commit()
        except:
            logging.exception(f"Saving pos failed: {line}")
            session.rollback()

        try:
            pairs = session.query(models.StreamDevices) \
                .filter(models.StreamDevices.device==device) \
                .all()
        except:
            logging.exception("Exception")
            session.rollback()
            return

        for pair in pairs:
            if not pair.stream:
                continue

            if not pair.stream.is_live:
                # We don't need to publish to record or proxy stream
                continue

            topic = f'{Constants.TOPIC_POS}.{pair.stream.serial}'
            await self.m_pub.send_multipart([
                topic.encode(),
                line.encode() ])


    async def handle_ipc_async(self, buf):
        try:
            obj = json.loads(buf)
        except Exception as e:
            logging.warning('JSON Exception ({}): {}'.format(type(e), e))
            return

        if obj['cmd'] == 'notify-datachanged':
            logging.info('DATA CHANGED NOTIFY')
            await self.flow.resume()


    def handle_ipc(self, buf):
        asyncio.create_task(self.handle_ipc_async(buf))


    async def watchdog(self):
        while True:
            await asyncio.sleep(30)
            await self.flow.resume()


    async def run_client(self):
        session = models.SessionSingleton.instance().session

        # Track to stop/remove devices
        devices_set = set()

        while True:
            logging.info("Checking pos devices")
            devices_set = set()

            try:
                devices: List[models.Device] = session.query(models.Device) \
                    .filter(models.Device.active==True) \
                    .all()
            except:
                logging.exception("Exception")
                session.rollback()
                break

            for device in devices:
                devices_set.add(device.name)

                if (device.address is None) or (device.port is None):
                    logging.warning(f"Device {device.name} has no address information, skipping")
                    continue

                try:
                    count = session.query(models.StreamDevices) \
                        .filter(models.StreamDevices.device==device) \
                        .count()
                except:
                    logging.exception("Exception")
                    session.rollback()
                    continue

                if int(count) == 0:
                    logging.debug(f"Device {device.name} has no associated stream, skipping")
                    continue

                if device.name in self.m_task_by_name:
                    info: DeviceInfo = self.m_task_by_name[device.name]
                    if info.task and (not info.task.done()) and info.last_updated==device.last_updated:
                        logging.debug(f"Device {device.name} no change, skipping")
                        continue

                    logging.warning(f"Device {device.name} changed or stopped")

                    # Device updated, cancelling current task
                    if info.task and (not info.task.done()):
                        logging.info('Canceling task for {}'.format(device.name))
                        info.task.cancel()
                        try:
                            await info.task
                        except asyncio.CancelledError:
                            pass

                logging.info(f"Creating task for {device.name}")
                task = asyncio.create_task(self.run_device(device))
                self.m_task_by_name[device.name] = DeviceInfo(task=task, last_updated=device.last_updated)

            for name in self.m_task_by_name:
                if name not in devices_set:
                    logging.warning(f"Device {name} is not available anymore, removing")

                    # Remove device since it is deleted
                    info: DeviceInfo = self.m_task_by_name[name]
                    if info.task and not info.task.done():
                        info.task.cancel()

                        try:
                            await info.task
                        except asyncio.CancelledError:
                            pass

                    del self.m_task_by_name[name]

            try:
                session.commit()
            except:
                logging.exception("Dummy commit failed")
                session.rollback()

            logging.info('Waiting for events')
            await self.flow.pause()


    async def run(self):
        self.setup()

        try:
            await asyncio.gather(
                self.run_ipc(),
                self.run_client(),
                self.watchdog())
        except:
            logging.exception("Pos client's tasks cancelled")

        self.cleanup()


def run_posclient():
    Config.init(board_cfg_file())
    setup_logger('pos_client.txt')

    config = Config.get()
    models.SessionSingleton.recreate(config.app.database)

    client = PosClient()

    try:
        asyncio.run(client.run())
    except:
        logging.info('Done')


if __name__ == '__main__':
    run_posclient()
