from typing import Dict, List, Any, Optional
import os
import re
import sys
import json
import uuid
import shlex
import logging
import asyncio
import datetime
import functools
import pathlib
from dataclasses import dataclass

from sqlalchemy import or_

import zmq
from zmq.asyncio import Context as ZmqContext
from zmq.asyncio import Socket as ZmqSocket

from package.shared.constants import Constants, StreamSetting
from package.shared.ffmpeg_parser import FFmpegParser
from package.shared import models
from package.shared.config import Config
from package.shared.logger import setup_logger
from package.shared.ipc import IPCWorker
from package.shared.util import board_cfg_file
from package.shared.flow_control import FlowControl
from package.board.tvt import App as TvtApp


@dataclass
class TaskInfo():
    task: Optional[asyncio.Task] = None
    stream: Optional[models.Stream] = None
    last_updated: datetime.datetime = None

    # Zmq
    pub: Optional[ZmqSocket] = None
    topic_video: Optional[str] = None

    # Monitor
    last_frame: str = None
    has_data: bool = None
    last_state: models.StreamActivity = models.StreamActivity.INIT

    # Record
    chunk: models.Chunk = None
    filename: str = None
    ffmpeg: asyncio.subprocess.Process = None


@dataclass
class VideoPublisher(IPCWorker):
    context: ZmqContext = None
    stream_flow: Optional[FlowControl] = None

    def __post_init__(self):
        super().__post_init__()


    def setup(self):
        config = Config.get()

        self.stream_flow = FlowControl()
        self.context = ZmqContext()

        # IPC
        name = 'vid_mgnt'
        url = config.app.ipc
        self.setup_ipc(
            url=url,
            ctx=self.context,
            type=zmq.DEALER,
            name=name)


    def cleanup(self):
        self.close_ipc()
        self.context.destroy()


    async def read_ffmpeg_stdout(self, info: TaskInfo, stdout):
        config = Config.get()
        bufsize = int(config.livestream.bufsize)

        while info.ffmpeg.returncode is None:
            buf = await stdout.read(bufsize)
            if not buf:
                break

            await info.pub.send_multipart([ info.topic_video, buf ], copy=False)


    async def read_ffmpeg_stderr(self, info: TaskInfo, stderr):
        metadata = True
        parser = FFmpegParser()
        session = models.SessionSingleton.instance().session
        stream = info.stream

        while info.ffmpeg.returncode is None:
            c = await stderr.read(1)
            if not c:
                break

            parser.append(c)
            if c != b'\r' and c != b'\n':
                continue

            chunk, meta, timeline = parser.process(meta=metadata, chunk=True, time=not info.has_data)

            if meta:
                stream.meta=meta
                try:
                    session.commit()
                except:
                    logging.exception(f'Update stream {stream.serial} metadata failed')
                    session.rollback()

                metadata = False
            
            if chunk:
                if info.chunk:
                    chunk_to_delete = None
                    chunk_to_add = None
                    try:
                        filesize = os.path.getsize(info.filename)
                        if filesize == 0:
                            raise ValueError(f'Filesize is zero ({info.filename}), skipping')
                    except OSError:
                        logging.warning(f'Chunk file not found: {info.filename}')
                        chunk_to_delete = info.chunk
                    except ValueError as e:
                        logging.warning(e)
                        chunk_to_delete = info.chunk
                    else:
                        info.chunk.completed = True
                        info.chunk.filesize = filesize
                        info.chunk.finished_at = datetime.datetime.now()
                        chunk_to_add = info.chunk

                    try:
                        if chunk_to_delete:
                            session.delete(chunk_to_delete)

                        if chunk_to_add:
                            session.add(chunk_to_add)

                        session.commit()
                    except:
                        logging.exception("Update chunk completetion failed")
                        session.rollback()

                    # Reset
                    info.chunk = None

                info.filename = chunk.strip()
                logging.info(f'New chunk at: {info.filename}')

                created_at = datetime.datetime.now()
                created_min = models.get_timestamp_min(created_at.timestamp())
                year = created_at.year
                month = created_at.month
                day = created_at.day

                info.chunk = models.Chunk(
                    completed=False,
                    created_at=created_at,
                    created_min=created_min,
                    year=year,
                    month=month,
                    day=day,
                    filename=info.filename,
                    filesize=0,
                    stream=info.stream)

                session.add(info.chunk)

                try:
                    pairs = session.query(models.StreamDevices) \
                        .filter(models.StreamDevices.stream==info.stream) \
                        .all()
                except:
                    logging.exception("Exception")
                    session.rollback()
                    pairs = []

                try:
                    for pair in pairs:
                        session.add(models.ChunkDevices(chunk=info.chunk, device=pair.device))

                    session.commit()
                except:
                    logging.warning("Adding new chunk failed")
                    session.rollback()

            if timeline:
                pattern = r'frame=\s*(.+?)\s'
                match = re.search(pattern, timeline)
                if match:
                    current_frame = match.group(1)
                    if (info.last_frame!=current_frame) and (not info.has_data):
                        info.has_data = True

                    info.last_frame = current_frame


    async def tvt_handler(self, info, stdin, op):
        info.has_data = True
        if info.last_frame is None:
            info.last_frame = 1
        else:
            info.last_frame += 1

        try:
            logging.info(f'Writing { len(op.data) } bytes')
            stdin.write(op.data)
            await stdin.drain()
        except:
            logging.exception('Failed to write ffmpeg stdin')
            raise


    async def tvt_async(self, info, stdin):
        from urllib.parse import urlparse

        url = info.stream.input_url
        o = urlparse(url)

        try:
            app = TvtApp(
                callback=functools.partial(self.tvt_handler, info, stdin),
                host=o.hostname,
                port=int(o.port),
                channel=int(o.path[1:]),
                username=o.username,
                password=o.password)

            await app.run()
        except:
            logging.exception(f'TVT quitted: { info.stream.serial }')

        if info.ffmpeg.returncode is None:
            info.ffmpeg.kill()


    async def ffmpeg_subprocess(self, info: TaskInfo, cmd: List[str]):
        stdin = None
        stdout = None

        if info.stream.klass.startswith('qsee'):
            stdin = asyncio.subprocess.PIPE

        if info.stream.is_forward:
            stdout = asyncio.subprocess.PIPE

        info.ffmpeg = await asyncio.create_subprocess_exec(*cmd,
            stdin=stdin,
            stdout=stdout,
            stderr=asyncio.subprocess.PIPE,
            limit=1024*1024)

        tasks = [ self.read_ffmpeg_stderr(info, info.ffmpeg.stderr) ]

        if info.stream.klass.startswith('qsee'):
            tasks.append(self.tvt_async(info, info.ffmpeg.stdin))

        if info.stream.is_forward:
            tasks.append(self.read_ffmpeg_stdout(info, info.ffmpeg.stdout))

        await asyncio.gather(*tasks)
        await info.ffmpeg.wait()
        logging.warning(f'FFMPEG for { info.stream.serial } finished')


    async def run_ffmpeg(self, info: TaskInfo):
        # Here, no MJPEG forward stream

        stream: models.Stream = info.stream
        session = models.SessionSingleton.instance().session
        config = Config.get()

        # FFMPEG
        # In params
        in_params = []

        klass = stream.klass
        if klass.startswith('mjpeg'):
            in_params.append('-use_wallclock_as_timestamps 1')
        elif klass.startswith('qsee'):
            in_params.append('-f h264')
            try:
                realtime_setting = session.query(models.Setting) \
                    .filter(
                        models.Setting.stream==stream,
                        models.Setting.key==StreamSetting.REALTIME) \
                    .first()


                if realtime_setting.value=='1' or realtime_setting.value.lower()=='true':
                    in_params.append('-re')
            except:
                pass
        else:
            try:
                tcp_setting = session.query(models.Setting) \
                    .filter(
                        models.Setting.stream==stream,
                        models.Setting.key==StreamSetting.USE_TCP) \
                    .first()

                if tcp_setting.value=='1' or tcp_setting.value.lower()=='true':
                    in_params.append('-rtsp_transport tcp')
            except:
                pass

            try:
                realtime_setting = session.query(models.Setting) \
                    .filter(
                        models.Setting.stream==stream,
                        models.Setting.key==StreamSetting.REALTIME) \
                    .first()

                if realtime_setting.value=='1' or realtime_setting.value.lower()=='true':
                    in_params.append('-re')
            except:
                pass

        # Out (forward) params
        out_params = '-vcodec copy'
        try:
            audio_setting = session.query(models.Setting) \
                .filter(
                    models.Setting.stream==stream,
                    models.Setting.key==StreamSetting.FORWARD_AUDIO) \
                .first()

            if audio_setting.value=='1' or audio_setting.value.lower()=='true':
                out_params = f'{out_params} -acodec copy'
        except:
            out_params = f'{out_params} -an'

        ext = 'ts'
        withVideo = '-vcodec copy'
        withAudio = '-acodec copy'

        klass = stream.klass
        serial = stream.serial

        # MJPEG record doesn't use input_url, mjpeg.py use it instead
        if klass.startswith('mjpeg'):
            url = 'http://127.0.0.1:{}/mjpeg?stream={}'.format(config.mjpeg.port, serial)
            ext = 'avi'
            withVideo = '-vcodec copy'
            withAudio = '-an'
        elif klass.startswith('qsee'):
            url = 'pipe:0'
            withAudio = '-an'
        else:
            url = stream.input_url

        ffmpeg = config.ffmpeg.location
        length = config.record.length
        location = config.record.location

        # Remove special characters in serial
        serial_quotes = { '/', '*', '?', ':', '%', '\\', '.', '>', '<', ',', '&', '!', '~', '`' }
        quoted_serial = serial
        for s in serial_quotes:
            quoted_serial = quoted_serial.replace(s, '_')

        try:
            pathlib.Path(f'{location}/{quoted_serial}').mkdir(parents=True, exist_ok=True)
        except:
            pass

        # Build commandline
        record_cmd = (f' {withVideo} {withAudio} '
            f'-f stream_segment '
            f'-segment_time {length} '
            f'-segment_atclocktime 1 '
            f'-reset_timestamps 1 '
            f'-strftime 1 "{location}/{quoted_serial}/%Y-%m-%d_%H-%M-%S.{ext}" ')

        forward_cmd = f' {out_params} -f mpegts -'

        if not stream.is_record:
            record_cmd = ''

        # Mjpeg use mjpeg.py
        if (not stream.is_forward) or (klass.startswith('mjpeg')):
            forward_cmd = ''

        cmdline = f'{ffmpeg} { " ".join(in_params) } -i "{url}"{record_cmd}{forward_cmd}'

        logging.info(f'Cmdline: {cmdline}')
        arguments = shlex.split(cmdline)

        restart = int(config.ffmpeg.restart)
        while True:
            info.last_frame = None
            info.has_data = False
            await self.ffmpeg_subprocess(info, arguments)

            logging.warning(f'Going to restart FFMPEG for { serial } in { restart } secs')
            await asyncio.sleep(restart)


    async def run_monitor(self, info: TaskInfo):
        config = Config.get()
        session = models.SessionSingleton.instance().session

        delay = int(config.livestream.monitor_time)
        down_count = 0
        stream: models.Stream = info.stream

        while True:
            await asyncio.sleep(delay)

            try:
                state = models.StreamActivity.UP if info.has_data else models.StreamActivity.DOWN
                if state != info.last_state:
                    logging.info(f"Stream state changed from { info.last_state } to { state } for { stream.serial }")
                    session.add(models.StreamHistory(
                        stream=stream,
                        time=datetime.datetime.now(),
                        activity=state.name))

                    info.last_state = state
                elif state == models.StreamActivity.DOWN:
                    down_count += 1
                    if down_count > 10:
                        if info.ffmpeg.returncode is None:
                            logging.warning(f'{stream.serial} SO MUCH DOWN, RESTARTING')
                            try:
                                info.ffmpeg.kill()
                            except:
                                logging.exception(f'Exception in killing ffmpeg for {stream.serial}')

                elif state == models.StreamActivity.UP:
                    down_count = 0

                if state == models.StreamActivity.DOWN:
                    logging.warning(f'Stream {stream.serial} down detected!')
                    continue

                info.has_data = False
                info.stream.last_activity = datetime.datetime.now()

                session.commit()
            except:
                logging.exception("Exception")
                session.rollback()


    async def start_stream(self, info: TaskInfo) -> None:
        logging.info(f'Starting stream: {info.stream.serial}')

        try:
            await asyncio.gather(
                self.run_ffmpeg(info),
                self.run_monitor(info))
        except asyncio.CancelledError:
            logging.warning(f'Stream {info.stream.serial} cancelled')
            raise
        except:
            logging.exception(f'Stream {info.stream.serial} got exception')


    def fill_task(self, stream: models.Stream) -> TaskInfo:
        config = Config.get()

        pub = None
        if stream.is_forward:
            pub = self.context.socket(zmq.PUB)
            pub.setsockopt(zmq.LINGER, 500)
            pub.setsockopt(zmq.TOS, 100)

            url = config.bridge.internal
            pub.connect(url)

        topic_video = '{}.{}'.format(Constants.TOPIC_LIVE, stream.serial).encode()
        info = TaskInfo(
            stream=stream,
            last_updated=stream.last_updated,
            pub=pub,
            topic_video=topic_video,
            last_frame=None,
            has_data=None,
            last_state=models.StreamActivity.INIT,
            chunk=None,
            filename=None)

        info.task = asyncio.create_task(self.start_stream(info))
        return info


    async def streams_manager(self) -> None:
        config = Config.get()

        session = models.SessionSingleton.instance().session
        gw_serial: str = config.app.gateway

        task_by_serial: Dict[str, TaskInfo] = dict()
        while True:
            try:
                gateway = session.query(models.Gateway) \
                    .filter(models.Gateway.serial==gw_serial) \
                    .first()
            except:
                logging.exception("Exception")
                session.rollback()
                await asyncio.sleep(5)
                continue

            if gateway is None:
                logging.warning(f'No gateway {gw_serial} available, wait for 15 secs')
                await asyncio.sleep(15)
                continue

            try:
                streams = session.query(models.Stream) \
                    .filter(
                        models.Stream.gateway==gateway,
                        or_(
                            models.Stream.is_record==True,
                            models.Stream.is_forward==True))
            except:
                logging.exception("Exception")
                session.rollback()
                await asyncio.sleep(5)
                continue

            for stream in streams:
                # This just a "forward" stream, we have mjpeg for this purpose, so just skip ...
                if stream.klass.startswith('mjpeg') and not stream.is_record:
                    logging.info(f'Skipping forward mjpeg stream {stream.serial}')
                    continue

                # New stream
                if stream.serial not in task_by_serial:
                    logging.info(f'Lauching stream: {stream.serial}')

                    task_by_serial[stream.serial] = self.fill_task(stream)
                    continue

                # Exist stream
                info = task_by_serial[stream.serial]
                if info.last_updated == stream.last_updated:
                    logging.info(f'Stream {stream.serial} no change so far, skipping ...')
                    continue

                logging.info(f'Stream {stream.serial} changed, updating ...')
                if info.task and not info.task.done():
                    info.task.cancel()
                    try:
                        await info.task
                    except asyncio.CancelledError:
                        logging.warning(f'  Stream {stream.serial} cancelled')

                task_by_serial[stream.serial] = self.fill_task(stream)

            try:
                session.commit()
            except:
                session.rollback()
                logging.exception("Dummy commit failed")

            logging.info('Waiting for events...')
            await self.stream_flow.pause()
            logging.info('Resuming ...')


    async def handle_ipc_async(self, buf):
        try:
            obj = json.loads(buf)
        except Exception as e:
            logging.warning('JSON Exception ({}): {}'.format(type(e), e))
            return

        if obj['cmd'] == 'notify-datachanged':
            logging.warning('DATA CHANGED NOTIFY')
            await self.stream_flow.resume()


    def handle_ipc(self, buf):
        asyncio.create_task(self.handle_ipc_async(buf))


    async def delay(self):
        while True:
            await asyncio.sleep(30)
            await self.stream_flow.resume()


    async def run(self):
        self.setup()

        try:
            await asyncio.gather(
                self.run_ipc(),
                self.streams_manager(),
                self.delay())
        except:
            logging.exception("vid_mgnt tasks exception")

        self.cleanup()


def run_vid_mgnt():
    Config.init(board_cfg_file())
    setup_logger('vid_mgnt.txt')

    config = Config.get()
    models.SessionSingleton.recreate(config.app.database)

    app = VideoPublisher()

    try:
        asyncio.run(app.run())
    except KeyboardInterrupt:
        logging.info('Done')


if __name__ == '__main__':
    run_vid_mgnt()
