import sys
import datetime
import logging
import base64
import binascii
import asyncio
from dataclasses import dataclass

from sqlalchemy import  and_

from starlette.authentication import (
    AuthenticationBackend, AuthenticationError, SimpleUser, UnauthenticatedUser,
    AuthCredentials, requires
)

from starlette.applications import Starlette
from starlette.endpoints import HTTPEndpoint
from starlette.responses import Response, JSONResponse, PlainTextResponse, StreamingResponse
from starlette.schemas import SchemaGenerator
from starlette.middleware.authentication import AuthenticationMiddleware

import uvicorn
import aiofiles
import dateutil
import dateutil.parser
import json

import rx
import rx.subject
from rx import operators as ops
from rx.scheduler.eventloop import AsyncIOScheduler


if __name__ == '__main__':
    import os
    sys.path.append(os.path.join(os.path.dirname(__file__), '../..'))

from package.shared import models
from package.shared.config import Config
from package.shared.ipc import IPCWorker
from package.shared.logger import setup_logger
from package.shared.flow_control import FlowControl
from package.shared.util import \
    web_get_streams, web_get_states, web_get_gateways, \
    web_search_chunks_by_pos, \
    board_cfg_file
from package.shared.handlers import ESDataHandler
from package.board.sync import do_sync

import package.shared.libass as libass


class BasicAuthBackend(AuthenticationBackend):
    async def authenticate(self, request):
        if "Authorization" not in request.headers:
            return

        auth = request.headers["Authorization"]
        try:
            scheme, credentials = auth.split()
            if scheme.lower() != 'basic':
                return

            decoded = base64.b64decode(credentials).decode("ascii")
        except (ValueError, UnicodeDecodeError, binascii.Error) as exc:
            raise AuthenticationError('Invalid basic auth credentials')

        username, _, password = decoded.partition(":")
        # TODO: Verify usename/password
        return AuthCredentials(["authenticated"]), SimpleUser(username)


def create_app(debug=True):
    app = Starlette(debug=debug)
    app.add_middleware(AuthenticationMiddleware, backend=BasicAuthBackend())

    @app.on_event('startup')
    async def startup_handler():
        pass


    @app.on_event('shutdown')
    async def shutdown_handler():
        pass


    @app.route('/')
    class Homepage(HTTPEndpoint):
        @requires('authenticated')
        async def get(self, request):
            return PlainTextResponse('Welcome from board')


    # Return years/months/days that contains recording chunks
    @app.route('/recordings')
    class RecordingsAvailability(HTTPEndpoint):
        async def get(self, request):
            try:
                serial = request.query_params['stream']
            except Exception as e:
                return Response(str(e), status_code=404, headers=None, media_type=None)

            session = models.SessionSingleton.instance().session

            try:
                stream = session.query(models.Stream).filter(models.Stream.serial==serial).first()
                if stream is None:
                    return Response('Stream not found', status_code=404, headers=None, media_type=None)

                rows = session.query(
                        models.Chunk.year,
                        models.Chunk.month,
                        models.Chunk.day) \
                    .filter(models.Chunk.stream==stream) \
                    .group_by(models.Chunk.year, models.Chunk.month, models.Chunk.day)
            except Exception as e:
                logging.exception("Exception")
                session.rollback()
                return Response(str(e), status_code=500, headers=None, media_type=None)

            result = dict()
            for r in rows:
                y = r.year
                m = r.month
                d = r.day

                if y not in result:
                    result[y] = dict()

                if m not in result[y]:
                    result[y][m] = []

                if d not in result[y][m]:
                    result[y][m].append(d)

            return JSONResponse(result)


    # Return list of available recodings (1 record/1 min)
    @app.route('/day')
    class RecordingsEndpoint(HTTPEndpoint):
        async def get(self, request):
            try:
                serial = request.query_params['stream']
                year = int(request.query_params['year'])
                month = int(request.query_params['month'])
                day = int(request.query_params['day'])
            except Exception as e:
                return Response(str(e), status_code=404, headers=None, media_type=None)

            session = models.SessionSingleton.instance().session

            try:
                stream = session.query(models.Stream).filter(models.Stream.serial==serial).first()
                if stream is None:
                    return Response('Stream not found', status_code=404, headers=None, media_type=None)

                time_begin = datetime.datetime(year=year,
                    month=month,
                    day=day)
                
                time_end = time_begin + datetime.timedelta(days=1)

                pairs = session.query(models.Chunk, models.Pos).filter(
                        models.Chunk.stream_id==stream.id,
                        models.Chunk.created_at>=time_begin,
                        models.Chunk.created_at<=time_end) \
                    .outerjoin(models.ChunkDevices, models.ChunkDevices.chunk_id==models.Chunk.id) \
                    .outerjoin(
                        models.Pos,
                        and_(
                            models.Pos.device_id==models.ChunkDevices.device_id,
                            models.Chunk.created_min==models.Pos.created_min)) \
                    .all()
            except Exception as e:
                logging.exception("Exception")
                session.rollback()
                return Response(str(e), status_code=500, headers=None, media_type=None)

            def format_result(pair):
                c = pair.Chunk
                p = pair.Pos
                obj = {
                    'id': c.id,
                    'created_at': c.created_at.isoformat(),
                    'timestamp_min': c.created_min,
                    'filesize': c.filesize,
                    'completed': c.completed,
                    'has_pos': True if p is not None else False
                }
                return obj

            objs = [ format_result(pair) for pair in pairs ]
            return JSONResponse(objs)


    # Return video data of streams in range
    @app.route('/chunks')
    class ChunksEndpoint(HTTPEndpoint):
        async def chunk_generator(self, stream, min_from, min_to):
            session = models.SessionSingleton.instance().session

            try:
                chunks = session.query(models.Chunk) \
                    .filter(
                        models.Chunk.stream==stream,
                        models.Chunk.created_min>=min_from,
                        models.Chunk.created_min<=min_to) \
                    .order_by(models.Chunk.created_at) \
                    .all()

                logging.info(f"Got {len(chunks)} chunks, start streaming")

                try:
                    for chunk in chunks:
                        filename = chunk.filename
                        async with aiofiles.open(filename, mode='rb') as f:
                            while True:
                                content = await f.read(188 * 10)
                                if not content:
                                    break

                                yield content
                except:
                    logging.exception('Loading chunk error')
                    return
            except:
                logging.exception("Exception")
                session.rollback()


        async def get(self, request):
            try:
                serial = request.query_params['stream']
                min_from = int(request.query_params['min_from'])
                min_to = int(request.query_params['min_to'])
            except Exception as e:
                return Response(str(e), status_code=404, headers=None, media_type=None)

            session = models.SessionSingleton.instance().session

            try:
                stream = session.query(models.Stream) \
                    .filter(models.Stream.serial==serial) \
                    .first()
            except Exception as e:
                logging.exception("Exception")
                session.rollback()
                return Response(str(e), status_code=500, headers=None, media_type=None)

            if stream is None:
                return Response('Stream not found', status_code=404, headers=None, media_type=None)

            return StreamingResponse(self.chunk_generator(stream, min_from, min_to), media_type='application/octet-stream')


    # Return chunks contains value in range (from - to) of a stream (or all streams)
    @app.route('/search')
    class SearchEndpoint(HTTPEndpoint):
        async def get(self, request):
            serial = None
            

            try:
                serial = request.query_params['stream']
            except:
                pass

            try:
                min_from = int(request.query_params['min_from'])
                min_to = int(request.query_params['min_to'])
                value = request.query_params['value']
            except Exception as e:
                return Response(str(e), status_code=404, headers=None, media_type=None)

            config = Config.get()
            session = models.SessionSingleton.instance().session

            try:
                gateway = session.query(models.Gateway) \
                    .filter(models.Gateway.serial==config.app.gateway) \
                    .first()
            except:
                logging.exception("Exception")
                session.rollback()
                return Response(str(e), status_code=500, headers=None, media_type=None)

            if gateway is None:
                return Response('Invalid gateway', status_code=404, headers=None, media_type=None)

            objs = web_search_chunks_by_pos(
                gateway,
                min_from,
                min_to,
                value,
                serial)

            return JSONResponse(objs)


    # Return subtitle
    @app.route('/sub')
    class SubEndpoint(HTTPEndpoint):
        async def sub_generator(self, serial, min_from, min_to, font):
            session = models.SessionSingleton.instance().session

            try:
                stream = session.query(models.Stream).filter(models.Stream.serial==serial).first()
                if stream is None:
                    logging.warning(f'No stream found: {serial}')
                    yield ''
                    return

                chunk = session.query(models.Chunk).filter(
                        models.Chunk.stream==stream,
                        models.Chunk.created_min>=min_from,
                        models.Chunk.created_min<=min_to) \
                    .order_by(models.Chunk.created_at) \
                    .first()

                if chunk is None:
                    logging.warning(f'No chunk in range: [{min_from}-{min_to}]')
                    yield ''
                    return

                # Get all poses in range
                rows = session.query(models.Pos) \
                    .join(models.ChunkDevices, models.ChunkDevices.device_id==models.Pos.device_id) \
                    .join(
                        models.Chunk,
                        and_(
                            models.Chunk.id==models.ChunkDevices.chunk_id,
                            models.Chunk.created_min==models.Pos.created_min,
                            models.Pos.created_at>=models.Chunk.created_at,
                            models.Pos.created_at<=models.Chunk.finished_at
                    )) \
                    .filter(
                        models.Chunk.stream==stream,
                        models.Chunk.created_min>=min_from,
                        models.Chunk.created_min<=min_to) \
                    .order_by(models.Pos.created_at) \
                    .all()
            except:
                logging.exception("Exception")
                session.rollback()
                return

            doc = libass.Document()

            size = font['size'] if 'size' in font else '15'
            color = font['color'] if 'color' in font else '&H00FFFFFF'
            outline = font['outline'] if 'outline' in font else '&H00000000'

            style = libass.Style.parse(f'Default,Arial,{size},{color},&H000000FF,{outline},&H00000000,0,0,0,0,100,100,0,0,1,1,0,7,10,10,10,1', doc.styles_field_order)
            doc.styles.append(style)

            start = chunk.created_at
            timeline = dict()

            for pos in rows:
                appear = pos.created_at - start

                second = round(appear.total_seconds())
                for i in range(0, 4):
                    time = second + i
                    if time not in timeline:
                        timeline[time] = []

                    timeline[time].append(pos.text)


            seconds = list(timeline.keys())
            seconds.sort()
            for second in seconds:
                for text in timeline[second]:
                    dlg = libass.Dialogue()
                    dlg.text = text
                    dlg.start = datetime.timedelta(seconds=second)
                    dlg.end = datetime.timedelta(seconds=second+1)
                    doc.events.append(dlg)

            import io
            strio = io.StringIO()
            doc.dump_file(strio)
            yield strio.getvalue()
            strio.close()


        async def get(self, request):
            try:
                serial = request.query_params['stream']
                min_from = int(request.query_params['min_from'])
                min_to = int(request.query_params['min_to'])
            except Exception as e:
                return Response(str(e), status_code=404, headers=None, media_type=None)

            return StreamingResponse(self.sub_generator(serial, min_from, min_to, request.query_params), media_type='text/plain')


    # List all POS Devices
    @app.route('/devices')
    class DevicesAvailability(HTTPEndpoint):
        async def get(self, request):
            session = models.SessionSingleton.instance().session

            try:
                devices = session.query(models.Device) \
                    .filter(models.Device.active==True) \
                    .all()
            except Exception as e:
                logging.exception("Exception")
                session.rollback()
                return Response(str(e), status_code=500, headers=None, media_type=None)

            def format_result(d):
                return {
                    'name': d.name,
                    'address': d.address,
                    'port': d.port
                }

            return JSONResponse([ format_result(d) for d in devices ])


    # List of gateways
    @app.route('/gateways')
    class GatewaysAvailability(HTTPEndpoint):
        async def get(self, request):
            return JSONResponse(web_get_gateways())


    # List all streams
    @app.route('/streams')
    class StreamsAvailability(HTTPEndpoint):
        async def get(self, request):
            config = Config.get()
            return JSONResponse(web_get_streams(config.app.gateway))


    # Stream states
    @app.route('/states')
    class StateEndpoint(HTTPEndpoint):
        async def get(self, request):
            config = Config.get()
            return JSONResponse(web_get_states(config.app.gateway))


    # History
    @app.route('/history')
    class HistoryEndpoint(HTTPEndpoint):
        async def get(self, request):
            try:
                serial = request.query_params['stream']
            except Exception as e:
                return Response(str(e), status_code=404, headers=None, media_type=None)

            session = models.SessionSingleton.instance().session

            try:
                history = session.query(models.StreamHistory) \
                    .filter(models.StreamHistory.stream.has(serial=serial)) \
                    .all()
            except:
                logging.exception("Exception")
                session.rollback()
                return Response(str(e), status_code=500, headers=None, media_type=None)

            if len(history) == 0:
                return Response('No history found', status_code=404, headers=None, media_type=None)

            def format_history(h):
                return {
                    'time': h.time.isoformat(),
                    'activity': h.activity
                }

            return JSONResponse([ format_history(h) for h in history ])


    # esdata
    @app.route('/esdata')
    class ESDataEndpoint(HTTPEndpoint):
        async def post(self, request):
            try:
                rq = await request.json()
                do_sync(rq)
            except:
                logging.exception('Failed to parse ES data')

            return JSONResponse(dict())

    return app


def run_web():
    Config.init(board_cfg_file())
    setup_logger('web.txt')

    config = Config.get()
    models.SessionSingleton.recreate(config.app.database)

    app = create_app(debug=True)
    uvicorn.run(app, host=config.web.host, port=int(config.web.port))


if __name__ == '__main__':
    run_web()
