from typing import Any
from enum import Enum
import datetime
import uuid
import os
import sys
import logging
from dataclasses import dataclass

from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, BigInteger, Integer, String, DateTime, Boolean, Text, ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.orm import sessionmaker, scoped_session

Base = declarative_base()

UNKNOW_ID = 1
UNKNOW = 'Unknow'


@dataclass
class SessionSingleton():
    url: str
    engine: Any = None
    Session: Any = None
    session: Any = None

    __instance = None

    @classmethod
    def instance(cls):
        if cls.__instance is None:
            raise RuntimeError('Instance is not initialized')

        return cls.__instance


    @classmethod
    def recreate(cls, url):
        cls.__instance = None
        SessionSingleton(url)

        return cls.__instance


    def __post_init__(self):
        if self.__class__.__instance is not None:
            raise ValueError('This is singleton!')

        self.__class__.__instance = self

        self.engine = create_engine(self.url, echo=False)
        self.Session = sessionmaker(bind=self.engine)
        self.session = self.Session()


class StreamActivity(Enum):
    INIT    = 'init'
    UP      = 'up'
    DOWN    = 'down'


def get_timestamp_min(timestamp):
    intvar = int(timestamp)
    return intvar - intvar % 60


async def cleanup():
    session = SessionSingleton.instance().session

    try:
        session.query(SMBFile).delete()
        session.query(Setting).delete()
        session.query(UserGroups).delete()
        session.query(Group).delete()
        session.query(User).delete()
        session.query(StreamHistory).delete()
        session.query(ChunkDevices).delete()
        session.query(Chunk).delete()
        session.query(Sensor).delete()
        session.query(Actuator).delete()
        session.query(StreamDevices).delete()
        session.query(Stream).delete()
        session.query(Gateway).delete()
        session.query(Pos).delete()
        session.query(Device).delete()

        session.commit()
    except:
        logging.exception("Clean db failed")
        session.rollback()


class Device(Base):
    __tablename__ = 'devices'

    id = Column(Integer, primary_key=True)
    name = Column(String(150), unique=True)
    address = Column(String(150), nullable=True)
    port = Column(Integer, nullable=True)
    active = Column(Boolean)
    last_checked = Column(DateTime, nullable=True, index=True)
    last_updated = Column(DateTime)

    poses = relationship('Pos', lazy='dynamic')


class Pos(Base):
    __tablename__ = 'pos'

    id = Column(Integer, primary_key=True)
    created_at = Column(DateTime, index=True)
    created_min = Column(Integer, index=True)
    text = Column(String(100000))
    device_id = Column(Integer, ForeignKey('devices.id', ondelete='CASCADE'), index=True)
    device = relationship(Device, back_populates='poses')


class Gateway(Base):
    __tablename__ = 'gateways'

    id = Column(Integer, primary_key=True)
    name = Column(String(150))
    serial = Column(String(150), unique=True, index=True)
    ldap_id = Column(String(150), unique=True)
    klass = Column(String(50))
    path = Column(String(1024))

    address = Column(String(150))
    port = Column(Integer)
    dataPort = Column(Integer)
    last_updated = Column(DateTime)
    last_checked = Column(DateTime, nullable=True, index=True)

    streams = relationship('Stream', lazy='dynamic')


class Stream(Base):
    __tablename__ = 'streams'

    id = Column(Integer, primary_key=True)
    name = Column(String(150))
    serial = Column(String(150), unique=True, index=True)
    ldap_id = Column(String(150), unique=True, index=True)
    klass = Column(String(50))
    path = Column(String(1024))

    gateway_id = Column(Integer, ForeignKey('gateways.id', ondelete='CASCADE'), index=True)
    gateway = relationship(Gateway, back_populates='streams')

    input_url = Column(String(2048), nullable=True)

    # Meta data (resolution, fps, bitrate, ...)
    meta = Column(String(2048), nullable=True)

    is_live = Column(Boolean, nullable=True)
    is_record = Column(Boolean, nullable=True)
    is_forward = Column(Boolean, nullable=True)
    zmq_port = Column(Integer, nullable=True)

    # Reference
    # Used in multiple viewports for same camera
    ref = Column(String(150), nullable=True)

    last_activity = Column(DateTime, nullable=True)
    last_updated = Column(DateTime)
    last_checked = Column(DateTime, nullable=True, index=True)


class StreamDevices(Base):
    __tablename__ = 'stream_devices'

    id = Column(Integer, primary_key=True)
    stream_id = Column(Integer, ForeignKey('streams.id', ondelete='CASCADE'), nullable=True, index=True)
    stream = relationship(Stream)
    device_id = Column(Integer, ForeignKey('devices.id', ondelete='CASCADE'), nullable=True, index=True)
    device = relationship(Device)
    last_updated = Column(DateTime)
    last_checked = Column(DateTime, nullable=True, index=True)


class Actuator(Base):
    __tablename__ = 'actuators'

    id = Column(Integer, primary_key=True)
    name = Column(String(150))
    serial = Column(String(150), unique=True)
    ldap_id = Column(String(150), unique=True)
    klass = Column(String(50))
    path = Column(String(1024))

    gateway_id = Column(Integer, ForeignKey('gateways.id', ondelete='CASCADE'), index=True)
    last_updated = Column(DateTime)
    last_checked = Column(DateTime, nullable=True, index=True)

    gateway = relationship(Gateway)


class Sensor(Base):
    __tablename__ = 'sensors'

    id = Column(Integer, primary_key=True)
    name = Column(String(150))
    serial = Column(String(150), unique=True)
    ldap_id = Column(String(150), unique=True)
    klass = Column(String(50))
    path = Column(String(1024))

    gateway_id = Column(Integer, ForeignKey('gateways.id', ondelete='CASCADE'), index=True)
    last_updated = Column(DateTime)
    last_checked = Column(DateTime, nullable=True, index=True)

    gateway = relationship(Gateway)


class Chunk(Base):
    __tablename__ = 'chunks'

    id = Column(Integer, primary_key=True)
    created_at = Column(DateTime, index=True)
    finished_at = Column(DateTime, nullable=True, index=True)
    created_min = Column(Integer, index=True)
    year = Column(Integer, index=True)
    month = Column(Integer, index=True)
    day = Column(Integer, index=True)
    stream_id = Column(Integer, ForeignKey('streams.id', ondelete='CASCADE'), index=True)
    filename = Column(String(1024))
    filesize = Column(BigInteger)
    completed = Column(Boolean)

    stream = relationship(Stream)


class ChunkDevices(Base):
    __tablename__ = 'chunk_devices'

    id = Column(Integer, primary_key=True)

    chunk_id = Column(Integer, ForeignKey('chunks.id', ondelete='CASCADE'), index=True)
    device_id = Column(Integer, ForeignKey('devices.id', ondelete='CASCADE'), index=True)

    chunk = relationship(Chunk)
    device = relationship(Device)


class StreamHistory(Base):
    __tablename__ = 'stream_history'

    id = Column(Integer, primary_key=True)
    stream_id = Column(Integer, ForeignKey('streams.id', ondelete='CASCADE'), index=True)
    time = Column(DateTime)
    activity = Column(String(150))

    stream = relationship(Stream)


class User(Base):
    __tablename__ = 'users'

    id = Column(Integer, primary_key=True)
    first_name = Column(String(100))
    last_name = Column(String(100))
    username = Column(String(50), unique=True)
    password = Column(String(150))
    last_updated = Column(DateTime)
    last_checked = Column(DateTime, nullable=True)


class Group(Base):
    __tablename__ = 'groups'

    id = Column(Integer, primary_key=True)
    name = Column(String(100))
    last_updated = Column(DateTime)
    last_checked = Column(DateTime, nullable=True)


class UserGroups(Base):
    __tablename__ = 'user_groups'

    id = Column(Integer, primary_key=True)
    user_id = Column(Integer, ForeignKey('users.id', ondelete='CASCADE'))
    group_id = Column(Integer, ForeignKey('groups.id', ondelete='CASCADE'))
    last_updated = Column(DateTime)
    last_checked = Column(DateTime, nullable=True)

    user = relationship(User)
    group = relationship(Group)


class Setting(Base):
    __tablename__ = 'settings'

    id = Column(Integer, primary_key=True)

    gateway_id = Column(Integer, ForeignKey('gateways.id', ondelete='CASCADE'), nullable=True, index=True)
    gateway = relationship(Gateway)

    stream_id = Column(Integer, ForeignKey('streams.id', ondelete='CASCADE'), nullable=True, index=True)
    stream = relationship(Stream)

    actuator_id = Column(Integer, ForeignKey('actuators.id', ondelete='CASCADE'), nullable=True, index=True)
    actuator = relationship(Actuator)

    sensor_id = Column(Integer, ForeignKey('sensors.id', ondelete='CASCADE'), nullable=True, index=True)
    sensor = relationship(Sensor)

    user_id = Column(Integer, ForeignKey('users.id', ondelete='CASCADE'), nullable=True, index=True)
    user = relationship(User)

    key = Column(String(150))
    value = Column(String(102400))

    last_updated = Column(DateTime, index=True)
    last_checked = Column(DateTime, nullable=True, index=True)


class SMBFile(Base):
    __tablename__ = 'smbs'

    id = Column(Integer, primary_key=True)
    filepath = Column(String(2048))
    filesize = Column(Integer)
    content = Column(Text)
    created = Column(String(100))
    last_modified = Column(String(100))
    retrieved_at = Column(DateTime)
    md5_sum = Column(String(50))
    last_checked = Column(DateTime, nullable=True)


class WholePos(Base):
    __tablename__ = 'wholepos'

    id = Column(Integer, primary_key=True)
    created_at = Column(DateTime, index=True)
    created_min = Column(Integer, index=True)
    text = Column(String(1000))
    device_id = Column(Integer, ForeignKey('devices.id', ondelete='CASCADE'), index=True)
    device = relationship(Device)


# Utilities
async def upsert_gateway(name:str, serial: str, ldap_id: str, path: str, address: str, port: int, dataPort: int):
    session = SessionSingleton.instance().session

    if isinstance(port, str):
        port = int(port)

    if isinstance(dataPort, str):
        dataPort = int(dataPort)

    now = datetime.datetime.now()
    name = name.upper()

    try:
        gateway = session.query(Gateway) \
            .filter(Gateway.serial==serial) \
            .first()

        if gateway is None:
            gateway = session.query(Gateway).filter(Gateway.ldap_id==ldap_id).first()
            if gateway:
                logging.info(f'Gateway serial changed from: { gateway.serial } to { serial }')


        if gateway is None:
            logging.info('Creating new gateway {}'.format(serial))
            gateway = Gateway(
                name=name,
                serial=serial,
                ldap_id=ldap_id,
                path=path,
                address=address,
                port=port,
                dataPort=dataPort,
                last_updated=now,
                last_checked=now)

            session.add(gateway)
        else:
            kwargs = dict()

            if name!=gateway.name:
                kwargs['name'] = name

            if serial!=gateway.serial:
                kwargs['serial'] = serial

            if ldap_id!=gateway.ldap_id:
                kwargs['ldap_id'] = ldap_id

            if path!=gateway.path:
                kwargs['path'] = path

            if address!=gateway.address:
                kwargs['address'] = address

            if port!=gateway.port:
                kwargs['port'] = port

            if dataPort!=gateway.dataPort:
                kwargs['dataPort'] = dataPort

            if len(kwargs) > 0:
                logging.info('Updating gateway {}:'.format(serial))
                logging.info(kwargs)

                kwargs['last_updated'] = now

            kwargs['last_checked'] = now
            for k, v in kwargs.items():
                setattr(gateway, k, v)

        session.commit()
    except:
        logging.exception("Upsert gateway failed")
        session.rollback()

    return gateway


async def upsert_user(first_name: str, last_name: str, username: str, password: str):
    session = SessionSingleton.instance().session

    now = datetime.datetime.now()

    try:
        user = session.query(User) \
            .filter(User.username==username) \
            .first()

        if user is None:
            logging.info('Creating new user: {}'.format(username))
            user = User(
                first_name=first_name,
                last_name=last_name,
                username=username,
                password=password,
                last_updated=now,
                last_checked=now)

            session.add(user)
        else:
            kwargs = dict()
            if first_name!=user.first_name:
                kwargs['first_name'] = first_name

            if last_name!=user.last_name:
                kwargs['last_name'] = last_name

            if password!=user.password:
                kwargs['password'] = password

            if len(kwargs) > 0:
                logging.info(f'Updating user: {username}')
                logging.info(kwargs)

                kwargs['last_updated'] = now

            kwargs['last_checked'] = now

            for k, v in kwargs.items():
                setattr(user, k, v)

        session.commit()
    except:
        logging.exception("Upsert user failed")
        session.rollback()

    return user


async def upsert_device(**kwargs):
    """
    Insert or update device (by name)
    
    @param name: Name of device (string)
    @param address: Host name (string)
    @param port: Port (int)
    """

    if 'name' not in kwargs:
        raise ValueError("'name' must be in kwargs")

    name = kwargs['name']
    del kwargs['name']

    session = SessionSingleton.instance().session
    now = datetime.datetime.now()

    updated = False
    device = None

    try:
        device = session.query(Device) \
            .filter(Device.name==name) \
            .first()

        if device is None:
            logging.info('Creating new Device: {}'.format(name))

            device = Device(
                **kwargs,
                name=name,
                active=True,
                last_updated=now,
                last_checked=now)

            session.add(device)
            updated = True
        else:
            if not device.active:
                device.active = True
                updated = True
            
            if len(kwargs)>0:
                updated = True
                for k,v in kwargs.items():
                    setattr(device, k, v)

            if updated:
                device.last_updated = now

            device.last_checked = now

        session.commit()
    except:
        logging.exception("Upsert device failed")
        session.rollback()
        updated = False

    return (updated, device)


async def upsert_stream_device(stream: Stream, device: Device):
    session = SessionSingleton.instance().session

    now = datetime.datetime.now()

    try:
        pair = session.query(StreamDevices) \
            .filter(
                StreamDevices.stream==stream,
                StreamDevices.device==device) \
            .first()

        if pair is None:
            pair = StreamDevices(
                stream=stream,
                device=device,
                last_updated=now,
                last_checked=now)
            
            session.add(pair)
        else:
            pair.last_checked=now

        session.commit()
    except:
        logging.exception("Upsert stream device association failed")
        session.rollback()

    return pair


async def upsert_stream(
        name: str,
        serial: str,
        ldap_id: str,
        path: str,
        klass: str,
        gateway: Gateway,
        input_url: str,
        is_live: Boolean,
        is_record: Boolean,
        is_forward: Boolean):

    session = SessionSingleton.instance().session
    now = datetime.datetime.now()

    try:
        stream = session.query(Stream) \
            .filter(Stream.serial==serial) \
            .first()

        if stream is None:
            stream = session.query(Stream) \
                .filter(Stream.ldap_id==ldap_id) \
                .first()

            if stream:
                logging.info(f'Stream serial changed from: {stream.serial} to {serial}')

        if stream is None:
            logging.info('Create new stream: {}'.format(serial))
            stream = Stream(
                name=name,
                serial=serial,
                ldap_id=ldap_id,
                path=path,
                klass=klass,
                gateway_id=gateway.id,
                input_url=input_url,
                is_live=is_live,
                is_record=is_record,
                is_forward=is_forward,
                last_updated=now,
                last_checked=now)

            session.add(stream)
        else:
            kwargs = dict()
            if name!=stream.name:
                kwargs['name'] = name

            if serial!=stream.serial:
                kwargs['serial'] = serial

            if ldap_id!=stream.ldap_id:
                kwargs['ldap_id'] = ldap_id

            if path!=stream.path:
                kwargs['path'] = path

            if klass!=stream.klass:
                kwargs['klass'] = klass

            if gateway.id!=stream.gateway.id:
                kwargs['gateway'] = gateway

            if input_url!=stream.input_url:
                kwargs['input_url'] = input_url

            if is_live!=stream.is_live:
                kwargs['is_live'] = is_live

            if is_record!=stream.is_record:
                kwargs['is_record'] = is_record

            if is_forward!=stream.is_forward:
                kwargs['is_forward'] = is_forward

            if len(kwargs) > 0:
                logging.info('Updating stream {}:'.format(serial))
                logging.info(kwargs)
            
                kwargs['last_updated'] = now

            kwargs['last_checked'] = now

            for k, v in kwargs.items():
                setattr(stream, k, v)

        session.commit()
    except:
        logging.exception("Upsert stream failed")
        session.rollback()

    return stream


async def upsert_gateway_setting(gateway: Gateway, key: str, value: str):
    session = SessionSingleton.instance().session
    now = datetime.datetime.now()

    try:
        pair = session.query(Setting) \
            .filter(Setting.gateway==gateway, Setting.key==key) \
            .first()

        if pair is None:
            pair = Setting(
                gateway_id=gateway.id,
                key=key,
                value=value,
                last_updated=now,
                last_checked=now)

            session.add(pair)
        else:
            kwargs = dict()
            if pair.value!=value:
                kwargs['value'] = value
                kwargs['last_updated'] = now

            kwargs['last_checked'] = now
            for k, v in kwargs.items():
                setattr(pair, k, v)

        session.commit()
    except:
        logging.exception("Upsert gateway settings failed")
        session.rollback()

    return pair


async def upsert_stream_setting(stream: Stream, key: str, value: str):
    session = SessionSingleton.instance().session
    now = datetime.datetime.now()

    try:
        pair = session.query(Setting) \
            .filter(Setting.stream==stream, Setting.key==key) \
            .first()

        if pair is None:
            pair = Setting(
                stream_id=stream.id,
                key=key,
                value=value,
                last_updated=now,
                last_checked=now)

            session.add(pair)
        else:
            kwargs = dict()
            if pair.value!=value:
                kwargs['value'] = value
                kwargs['last_updated'] = now

            kwargs['last_checked'] = now
            for k, v in kwargs.items():
                setattr(pair, k, v)

        session.commit()
    except:
        logging.exception("Upsert stream settings failed")
        session.rollback()

    return pair


async def delete_users_older_than(val):
    session = SessionSingleton.instance().session

    try:
        session.query(User) \
            .filter(User.last_checked<val) \
            .delete()

        session.commit()
    except:
        logging.exception("Clean expired users failed")
        session.rollback()


async def delete_devices_older_than(val):
    session = SessionSingleton.instance().session

    try:
        session.query(Device) \
            .filter(Device.last_checked<val) \
            .update({ Device.active: False})

        session.commit()
    except:
        logging.exception("Clean expired devices failed")
        session.rollback()


async def delete_stream_settings_older_than(stream, val):
    session = SessionSingleton.instance().session

    try:
        session.query(Setting) \
            .filter(Setting.stream==stream, Setting.last_checked<val) \
            .delete()

        session.commit()
    except:
        logging.exception("Clean expired settings failed")
        session.rollback()


async def delete_stream_devices_older_than(stream, val):
    session = SessionSingleton.instance().session

    try:
        query = session.query(StreamDevices) \
            .filter(StreamDevices.last_checked<val)

        if stream:
            query = query \
                .filter(StreamDevices.stream_id==stream.id)

        query.delete()
        session.commit()
    except:
        logging.exception("Clean expired devices failed")
        session.rollback()


async def delete_streams_older_than(gateway, val):
    session = SessionSingleton.instance().session

    try:
        session.query(Stream) \
            .filter(Stream.gateway==gateway, Stream.last_checked<val) \
            .delete()

        session.commit()
    except:
        logging.exception("Delete stream failed")
        session.rollback()


async def delete_gateways_older_than(val):
    session = SessionSingleton.instance().session

    try:
        session.query(Gateway) \
            .filter(Gateway.last_checked<val) \
            .delete()

        session.commit()
    except:
        logging.exception("Clean expired gateways failed")
        session.rollback()
