import sys
import os
import logging
import uuid
import datetime
from typing import Dict, List, Optional

import json
import aiohttp
from sqlalchemy import and_

from package.shared.constants import ESObjectType, StreamSetting, GatewaySetting, Stream_DataReplacement
import package.shared.models as models


def base_dir():
    return os.path.abspath(os.path.join(
        os.path.dirname(__file__),
        '../..'))


def board_cfg_file():
    return '/etc/newsvc/conf/board.yaml'


def server_cfg_file():
    return '/etc/newsvc/conf/server.yaml'


def parse_url(url: str):
    for k, v in Stream_DataReplacement.items():
        url = url.replace(k, v)

    return url


def parse_es_result(reply, Val, Obj):
    serials_set = set()
    obj_by_id = dict()
    obj_path_by_id = dict()

    for obj in reply.obj:
        obj_by_id[obj.id] = obj
        if hasattr(obj, 'serial') and obj.serial in serials_set:
            logging.warning(f"Serial duplicated: {obj.serial} ({obj.name})")

    for obj_id, obj in obj_by_id.items():
        current = obj_by_id[obj.parent] if obj.parent in obj_by_id else None
        path = []

        while current:
            path.insert(0, current.name)

            if current.parent in obj_by_id:
                current = obj_by_id[current.parent]
            else:
                current = None

        full_path = '/'.join(path)
        if not full_path.startswith('/'):
            full_path = f"/{full_path}"

        obj_path_by_id[obj_id] = full_path

    # TODO:
    # Validation:
    # -  serial must exists for dev/strm/actr/sens and unique
    # -  url must exist in stream


    def obj_settings(obj):
        settings = None

        for c in obj.conf:
            name = c.name
            cval = c.val

            val = None
            if cval.type == Val.STR:
                val = cval.str_val
            elif cval.type == Val.INT:
                val = cval.int_val
            else:
                logging.warning(f'Unsupported settings type: {cval.type}')
                continue

            if settings is None:
                settings = dict()

            settings[name.lower()] = str(val)

        return settings


    def parse_stream(strm):
        name = strm.name
        serial = strm.strm_obj.serial_num
        if not serial:
            logging.warning(f"Stream {strm.name} has no serial, skip")
            serial = str(uuid.uuid4())

        ldap_id = str(strm.id)
        klass = strm.cls.lower()

        is_live = klass.endswith('.live')
        is_forward = klass.endswith('.proxy')
        is_record = klass.endswith('.proxy')

        settings = obj_settings(strm)
        if settings is None:
            logging.warning(f'Stream {serial} has no settings, skip')
            return None

        if StreamSetting.URL not in settings:
            logging.warning(f'No URL setting in stream: {serial}, skip')
            return None

        input_url = settings[StreamSetting.URL].strip()
        if len(input_url)==0:
            logging.warning(f'No stream URL is empty: {serial}, skip')
            return None

        input_url = parse_url(input_url)

        # Record
        if StreamSetting.NO_RECORD in settings:
            val = str(settings[StreamSetting.NO_RECORD]).lower()
            if val=='1' or val=='true':
                is_record = False

        # Forward
        if StreamSetting.NO_FORWARD in settings:
            val = str(settings[StreamSetting.NO_FORWARD]).lower()
            if val=='1' or val=='true':
                is_forward = False

        poses = []
        if StreamSetting.POS in settings:
            poses = [ s.strip().upper() for s in settings[StreamSetting.POS].split(',') if len(s.strip())>0 ]

        # Remove processed settings
        settings.pop(StreamSetting.URL, None)
        settings.pop(StreamSetting.POS, None)
        settings.pop(StreamSetting.NO_RECORD, None)
        settings.pop(StreamSetting.NO_FORWARD, None)

        return {
            'name': name,
            'serial': serial,
            'ldap_id': ldap_id,
            'klass': klass,
            'path': obj_path_by_id[strm.id],
            'input_url': input_url,
            'is_live': is_live,
            'is_record': is_record,
            'is_forward': is_forward,
            'poses': poses,
            'settings': settings
        }


    def parse_gateway(gateway):
        name = gateway.name
        serial = gateway.dev_obj.serial_num
        if not serial:
            serial = str(uuid.uuid4())

        ldap_id = str(gateway.id)
        klass = gateway.cls.lower()

        settings = obj_settings(gateway)
        if settings is None:
            settings = dict()

        address = '127.0.0.1'
        if GatewaySetting.HOST in settings:
            address = settings[GatewaySetting.HOST]

        port = '8000'
        if GatewaySetting.PORT in settings:
            port = settings[GatewaySetting.PORT]

        dataPort = '5560'
        if GatewaySetting.DATA_PORT in settings:
            dataPort = settings[GatewaySetting.DATA_PORT]

        return {
            'name': name,
            'serial': serial,
            'ldap_id': ldap_id,
            'klass': klass,
            'path': obj_path_by_id[gateway.id],
            'address': address,
            'port': port,
            'dataPort': dataPort
        }


    def parse_user(user):
        first_name = user.user_obj.first_name
        last_name = user.user_obj.last_name
        username = user.user_obj.user_name or str(user.id)

        return {
            'first_name': first_name,
            'last_name': last_name,
            'username': username,
            'password': 'N/A'
        }

    # Result
    gateway_by_id = dict()
    streams_by_gateway = dict()
    user_streams = dict()
    users = []

    # Poses
    poses_set = set()

    # received objects
    for obj_id, obj in obj_by_id.items():
        logging.info("object id: {0}, name: {1}".format(obj_id, obj.name))

        if obj.type == Obj.DEV:
            gateway_by_id[obj.id] = parse_gateway(obj)
        elif obj.type == Obj.STRM:
            if obj.parent not in streams_by_gateway:
                streams_by_gateway[obj.parent] = []

            strm = parse_stream(obj)
            if strm is None:
                continue

            streams_by_gateway[obj.parent].append(strm)
            for p in strm['poses']:
                poses_set.add(p)

            if obj.parent not in user_streams:
                user_streams[obj.parent] = []

            user_streams[obj.parent].append(strm['serial'])
        elif obj.type == Obj.USER:
            user = parse_user(obj)
            users.append(user)


    gateways = []
    for gw_id, gw in gateway_by_id.items():
        gw['streams'] = []
        if gw_id in streams_by_gateway:
            for strm in streams_by_gateway[gw_id]:
                gw['streams'].append(strm)

        gateways.append(gw)

    result = {
        'gateways': gateways,
        'users': users,
        'user_streams': user_streams,
        'poses': list(poses_set)
    }

    return result


def web_get_gateways():
    session = models.SessionSingleton.instance().session

    try:
        gateways = session.query(models.Gateway)
    except:
        logging.exception("Exception")
        session.rollback()
        return []

    def format_output(s):
        return {
            'name': s.name,
            'serial': s.serial,
            'ldap_id': s.ldap_id,
            'address': s.address,
            'port': s.port,
            'dataPort': s.dataPort
        }

    return [ format_output(gw) for gw in gateways ]


def web_get_streams(gwserial=None):
    session = models.SessionSingleton.instance().session

    try:
        gateways = session.query(models.Gateway)

        if gwserial:
            gateways = gateways.filter(models.Gateway.serial==gwserial)
    except:
        logging.exception("Exception")
        session.rollback()
        return []

    def format_stream(s):
        settings = dict()
        rows = session.query(models.Setting) \
            .filter(models.Setting.stream==s) \
            .all()

        for setting in rows:
            settings[setting.key] = setting.value

        devices = []
        pairs = session.query(models.StreamDevices) \
            .filter(models.StreamDevices.stream==s) \
            .all()

        for pair in pairs:
            devices.append(pair.device.name)

        return {
            'name': s.name,
            'serial': s.serial,
            'url': s.input_url,
            'settings': settings,
            'devices': devices,
            'is_live': s.is_live,
            'is_record': s.is_record,
            'is_forward': s.is_forward,
            'meta': s.meta,
            'last_activity': s.last_activity.isoformat() if s.last_activity else None
        }

    def format_obj(s):
        return {
            'name': s.name,
            'serial': s.serial
        }

    objs = []
    for gateway in gateways:
        streams = session.query(models.Stream) \
            .filter(models.Stream.gateway==gateway) \
            .all()

        actrs = session.query(models.Actuator) \
            .filter(models.Actuator.gateway==gateway) \
            .all()

        sens = session.query(models.Sensor) \
            .filter(models.Sensor.gateway==gateway) \
            .all()

        objs.append({
            'name': gateway.name,
            'serial': gateway.serial,
            'address': gateway.address,
            'port': gateway.port,
            'dataPort': gateway.dataPort,
            'streams': [ format_stream(s) for s in streams ],
            'actuators': [ format_obj(s) for s in actrs ],
            'sensors': [ format_obj(s) for s in sens ],
        })

    return objs


def web_get_states(gwserial=None):
    session = models.SessionSingleton.instance().session

    streams = session.query(models.Stream).filter(models.Stream.is_live==False)
    if gwserial:
        gateway = session.query(models.Gateway).filter(models.Gateway.serial==gwserial).first()
        if not gateway:
            return {}

        streams = streams.filter(models.Stream.gateway_id==gateway.id)

    obj = dict()

    now = datetime.datetime.now()
    for stream in streams:
        gw = stream.gateway
        if not gw.name in obj:
            obj[gw.name] = dict()

        is_active = (now - stream.last_activity).total_seconds() < 15 if stream.last_activity else False
        obj[gw.name][stream.serial] = is_active

    return obj


def web_search_chunks_by_pos(
        gateway: models.Gateway,
        min_from: int,
        min_to: int,
        value: str,
        serial: Optional[str] = None):

    session = models.SessionSingleton.instance().session
    query = session.query(models.Chunk, models.Pos, models.Stream, models.Gateway) \
        .join(models.ChunkDevices, models.ChunkDevices.chunk_id==models.Chunk.id) \
        .join(models.Stream, models.Stream.id==models.Chunk.stream_id) \
        .join(models.Gateway, models.Gateway.id==models.Stream.gateway_id) \
        .join(
            models.Pos,
            and_(
                models.Pos.device_id==models.ChunkDevices.device_id,
                models.Pos.created_min==models.Chunk.created_min,
                models.Pos.created_at>=models.Chunk.created_at,
                models.Pos.created_at<=models.Chunk.finished_at)
        ) \
        .filter(
            models.Gateway.id==gateway.id,
            models.Chunk.created_min>=min_from,
            models.Chunk.created_min<=min_to,
            models.Pos.text.ilike(f'%{value}%')) \

    if serial:
        query = query.filter(models.Stream.serial==serial)

    query.order_by(
        models.Stream.id,
        models.Chunk.created_at)

    rows = query.all()

    def format_result(row):
        gateway = row.Gateway
        stream = row.Stream
        chunk = row.Chunk
        pos = row.Pos

        return {
            'gw_address': gateway.address,
            'gw_port': gateway.port,
            'gw_dataport': gateway.dataPort,
            'strm_serial': stream.serial,
            'strm_name': stream.name,
            'created_at': chunk.created_at.isoformat(),
            'finished_at': chunk.finished_at.isoformat(),
            'created_min': chunk.created_min,
            'completed': chunk.completed,
            'pos_text': pos.text
        }

    objs = [ format_result(row) for row in rows ]
    return objs


async def get_request_json(url: str):
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            return await response.json()


async def post_request_json(url: str, obj: Dict):
    async with aiohttp.ClientSession() as session:
        async with session.post(url, json=obj) as response:
            return await response.json()